/*
 * kgraft_patch_bsc1189420
 *
 * Fix for CVE-2021-3653, bsc#1189420
 *
 *  Upstream commit:
 *  0f923e07124d ("KVM: nSVM: avoid picking up unsupported bits from L2 in
 *                 int_ctl (CVE-2021-3653)")
 *
 *  SLE12-SP3 commit:
 *  0f83408195dd943fde0b6a63e8c962af157a3627
 *
 *  SLE12-SP4, SLE15 and SLE15-SP1 commit:
 *  9c35f8d4e13594b6cf0357e3d5aa5f4dbdf379b7
 *
 *  SLE12-SP5 commits:
 *  9c35f8d4e13594b6cf0357e3d5aa5f4dbdf379b7
 *  a1c39b14d6b74df84321ae691c82b622f343e68a
 *
 *  SLE15-SP2 and -SP3 commit:
 *  790261594034490f018538690ab890fad71a62f4
 *
 *
 *  Copyright (c) 2021 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if IS_ENABLED(CONFIG_KVM_AMD)

#if !IS_MODULE(CONFIG_KVM_AMD)
#error "Live patch supports only CONFIG_KVM_AMD=m"
#endif

/* from include/linux/tracepoint.h */
#define KGRR___DECLARE_TRACE(name, proto, args, cond, data_proto, data_args) \
	static struct tracepoint (*kgre___tracepoint_##name);		\
	static inline void kgrr_trace_##name(proto)			\
	{								\
	if (unlikely(static_key_enabled(&(*kgre___tracepoint_##name).key))) \
		__DO_TRACE(&(*kgre___tracepoint_##name),		\
				TP_PROTO(data_proto),			\
				TP_ARGS(data_args),			\
				TP_CONDITION(cond),,);			\
		if (IS_ENABLED(CONFIG_LOCKDEP) && (cond)) {		\
			rcu_read_lock_sched_notrace();			\
			rcu_dereference_sched((*kgre___tracepoint_##name).funcs); \
			rcu_read_unlock_sched_notrace();		\
		}							\
	}

#define KGRR_DECLARE_TRACE(name, proto, args)				\
	KGRR___DECLARE_TRACE(name, PARAMS(proto), PARAMS(args),		\
			cpu_online(raw_smp_processor_id()),		\
			PARAMS(void *__data, proto),			\
			PARAMS(__data, args))

#define KGRR_TRACE_EVENT(name, proto, args)			\
	KGRR_DECLARE_TRACE(name, PARAMS(proto), PARAMS(args))



/* klp-ccp: from arch/x86/kvm/svm.c */
#define pr_fmt(fmt) "SVM: " fmt

#include <linux/kvm_host.h>

/* klp-ccp: from arch/x86/include/asm/kvm_host.h */
static struct kvm_x86_ops *(*kgre_kvm_x86_ops);

static void (*kgre_kvm_mmu_reset_context)(struct kvm_vcpu *vcpu);

static int (*kgre_kvm_set_cr3)(struct kvm_vcpu *vcpu, unsigned long cr3);

static unsigned long (*kgre_kvm_get_rflags)(struct kvm_vcpu *vcpu);
static void (*kgre_kvm_set_rflags)(struct kvm_vcpu *vcpu, unsigned long rflags);

static void (*kgre_kvm_mmu_unload)(struct kvm_vcpu *vcpu);

/* klp-ccp: from include/linux/kvm_host.h */
static void (*kgre_kvm_release_page_dirty)(struct page *page);

static int (*kgre_kvm_vcpu_read_guest)(struct kvm_vcpu *vcpu, gpa_t gpa, void *data,
			unsigned long len);

/* klp-ccp: from arch/x86/kvm/irq.h */
#include <linux/mm_types.h>
#include <linux/hrtimer.h>
#include <linux/kvm_host.h>
#include <linux/spinlock.h>

/* klp-ccp: from arch/x86/kvm/ioapic.h */
#include <linux/kvm_host.h>
#include <kvm/iodev.h>
/* klp-ccp: from arch/x86/kvm/lapic.h */
#include <kvm/iodev.h>
#include <linux/kvm_host.h>
/* klp-ccp: from arch/x86/kvm/mmu.h */
#include <linux/kvm_host.h>

/* klp-ccp: from arch/x86/kvm/kvm_cache_regs.h */
#define KVM_POSSIBLE_CR0_GUEST_BITS X86_CR0_TS

static inline unsigned long kgrr_kvm_register_read(struct kvm_vcpu *vcpu,
					      enum kvm_reg reg)
{
	if (!test_bit(reg, (unsigned long *)&vcpu->arch.regs_avail))
		(*kgre_kvm_x86_ops)->cache_reg(vcpu, reg);

	return vcpu->arch.regs[reg];
}

static inline void kvm_register_write(struct kvm_vcpu *vcpu,
				      enum kvm_reg reg,
				      unsigned long val)
{
	vcpu->arch.regs[reg] = val;
	__set_bit(reg, (unsigned long *)&vcpu->arch.regs_dirty);
	__set_bit(reg, (unsigned long *)&vcpu->arch.regs_avail);
}

static inline unsigned long kgrr_kvm_rip_read(struct kvm_vcpu *vcpu)
{
	return kgrr_kvm_register_read(vcpu, VCPU_REGS_RIP);
}

static inline void kvm_rip_write(struct kvm_vcpu *vcpu, unsigned long val)
{
	kvm_register_write(vcpu, VCPU_REGS_RIP, val);
}

static inline ulong kgrr_kvm_read_cr0_bits(struct kvm_vcpu *vcpu, ulong mask)
{
	ulong tmask = mask & KVM_POSSIBLE_CR0_GUEST_BITS;
	if (tmask & vcpu->arch.cr0_guest_owned_bits)
		(*kgre_kvm_x86_ops)->decache_cr0_guest_bits(vcpu);
	return vcpu->arch.cr0 & mask;
}

static inline ulong kgrr_kvm_read_cr0(struct kvm_vcpu *vcpu)
{
	return kgrr_kvm_read_cr0_bits(vcpu, ~0UL);
}

static inline ulong kgrr_kvm_read_cr3(struct kvm_vcpu *vcpu)
{
	if (!test_bit(VCPU_EXREG_CR3, (ulong *)&vcpu->arch.regs_avail))
		(*kgre_kvm_x86_ops)->decache_cr3(vcpu);
	return vcpu->arch.cr3;
}

static inline void enter_guest_mode(struct kvm_vcpu *vcpu)
{
	vcpu->arch.hflags |= HF_GUEST_MASK;
}

static inline bool is_guest_mode(struct kvm_vcpu *vcpu)
{
	return vcpu->arch.hflags & HF_GUEST_MASK;
}

/* klp-ccp: from arch/x86/kvm/mmu.h */
#define PT64_ROOT_LEVEL 4

static void
(*kgre_reset_shadow_zero_bits_mask)(struct kvm_vcpu *vcpu, struct kvm_mmu *context);

static void (*kgre_kvm_init_shadow_mmu)(struct kvm_vcpu *vcpu);

/* klp-ccp: from arch/x86/kvm/x86.h */
#include <asm/processor.h>
#include <linux/kvm_host.h>

static inline void kvm_clear_exception_queue(struct kvm_vcpu *vcpu)
{
	vcpu->arch.exception.pending = false;
}

static inline void kvm_clear_interrupt_queue(struct kvm_vcpu *vcpu)
{
	vcpu->arch.interrupt.pending = false;
}

static inline bool mmu_is_nested(struct kvm_vcpu *vcpu)
{
	return vcpu->arch.walk_mmu == &vcpu->arch.nested_mmu;
}

/* klp-ccp: from arch/x86/kvm/svm.c */
#include <linux/mod_devicetable.h>
#include <linux/kernel.h>
#include <linux/highmem.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <asm/apic.h>
#include <asm/perf_event.h>
#include <asm/tlbflush.h>
#include <asm/desc.h>
#include <asm/kvm_para.h>
#include <asm/virtext.h>

/* klp-ccp: from arch/x86/kvm/trace.h */
#include <linux/tracepoint.h>
#include <asm/vmx.h>
#include <asm/svm.h>
#include <asm/clocksource.h>
#include <asm/pvclock-abi.h>

KGRR_TRACE_EVENT(kvm_nested_vmrun,
	    TP_PROTO(__u64 rip, __u64 vmcb, __u64 nested_rip, __u32 int_ctl,
		     __u32 event_inj, bool npt),
	    TP_ARGS(rip, vmcb, nested_rip, int_ctl, event_inj, npt)
);

KGRR_TRACE_EVENT(kvm_nested_intercepts,
	    TP_PROTO(__u16 cr_read, __u16 cr_write, __u32 exceptions, __u64 intercept),
	    TP_ARGS(cr_read, cr_write, exceptions, intercept)
);

/* klp-ccp: from arch/x86/kvm/svm.c */
static const u32 host_save_user_msrs[10];

#define NR_HOST_SAVE_USER_MSRS ARRAY_SIZE(host_save_user_msrs)

struct nested_state {
	struct vmcb *hsave;
	u64 hsave_msr;
	u64 vm_cr_msr;
	u64 vmcb;

	/* These are the merged vectors */
	u32 *msrpm;

	/* gpa pointers to the real vectors */
	u64 vmcb_msrpm;
	u64 vmcb_iopm;

	/* A VMEXIT is required but not yet emulated */
	bool exit_required;

	/* cache for intercepts of the guest */
	u32 intercept_cr;
	u32 intercept_dr;
	u32 intercept_exceptions;
	u64 intercept;

	/* Nested Paging related state */
	u64 nested_cr3;
};

#define MSRPM_OFFSETS	16
static u32 (*kgre_msrpm_offsets)[MSRPM_OFFSETS] __read_mostly;

struct vcpu_svm {
	struct kvm_vcpu vcpu;
	struct vmcb *vmcb;
	unsigned long vmcb_pa;
	struct svm_cpu_data *svm_data;
	uint64_t asid_generation;
	uint64_t sysenter_esp;
	uint64_t sysenter_eip;
	uint64_t tsc_aux;

	u64 next_rip;

	u64 host_user_msrs[NR_HOST_SAVE_USER_MSRS];
	struct {
		u16 fs;
		u16 gs;
		u16 ldt;
		u64 gs_base;
	} host;

	u64 spec_ctrl;
	/*
	 * Contains guest-controlled bits of VIRT_SPEC_CTRL, which will be
	 * translated into the appropriate L2_CFG bits on the host to
	 * perform speculative control.
	 */
	u64 virt_spec_ctrl;

	u32 *msrpm;

	ulong nmi_iret_rip;

	struct nested_state nested;

	bool nmi_singlestep;

	unsigned int3_injected;
	unsigned long int3_rip;
	u32 apf_reason;

	/* cached guest cpuid flags for faster access */
	bool nrips_enabled	: 1;

	u32 ldr_reg;
	u32 dfr_reg;
	struct page *avic_backing_page;
	u64 *avic_physical_id_cache;
	bool avic_is_running;

	/*
	 * Per-vcpu list of struct amd_svm_iommu_ir:
	 * This is used mainly to store interrupt remapping information used
	 * when update the vcpu affinity. This avoids the need to scan for
	 * IRTE and try to match ga_tag in the IOMMU driver.
	 */
	struct list_head ir_list;
	spinlock_t ir_list_lock;
};

#if defined(CONFIG_X86_64) || defined(CONFIG_X86_PAE)
static bool (*kgre_npt_enabled);
#else
#error "klp-ccp: non-taken branch"
#endif

static void (*kgre_svm_set_cr0)(struct kvm_vcpu *vcpu, unsigned long cr0);
static void (*kgre_svm_flush_tlb)(struct kvm_vcpu *vcpu);

static int (*kgre_nested_svm_vmexit)(struct vcpu_svm *svm);

static inline void mark_all_dirty(struct vmcb *vmcb)
{
	vmcb->control.clean = 0;
}

static void (*kgre_recalc_intercepts)(struct vcpu_svm *svm);

static inline struct vmcb *get_host_vmcb(struct vcpu_svm *svm)
{
	if (is_guest_mode(&svm->vcpu))
		return svm->nested.hsave;
	else
		return svm->vmcb;
}

static inline void kgrr_clr_cr_intercept(struct vcpu_svm *svm, int bit)
{
	struct vmcb *vmcb = get_host_vmcb(svm);

	vmcb->control.intercept_cr &= ~(1U << bit);

	(*kgre_recalc_intercepts)(svm);
}

static inline void kgrr_clr_intercept(struct vcpu_svm *svm, int bit)
{
	struct vmcb *vmcb = get_host_vmcb(svm);

	vmcb->control.intercept &= ~(1ULL << bit);

	(*kgre_recalc_intercepts)(svm);
}

static inline void enable_gif(struct vcpu_svm *svm)
{
	svm->vcpu.arch.hflags |= HF_GIF_MASK;
}

static int get_npt_level(void)
{
#ifdef CONFIG_X86_64
	return PT64_ROOT_LEVEL;
#else
#error "klp-ccp: non-taken branch"
#endif
}

static void (*kgre_svm_set_efer)(struct kvm_vcpu *vcpu, u64 efer);

static void (*kgre_svm_set_cr0)(struct kvm_vcpu *vcpu, unsigned long cr0);

static int (*kgre_svm_set_cr4)(struct kvm_vcpu *vcpu, unsigned long cr4);

static unsigned long (*kgre_nested_svm_get_tdp_cr3)(struct kvm_vcpu *vcpu);

static u64 (*kgre_nested_svm_get_tdp_pdptr)(struct kvm_vcpu *vcpu, int index);

static void (*kgre_nested_svm_set_tdp_cr3)(struct kvm_vcpu *vcpu,
				   unsigned long root);

static void (*kgre_nested_svm_inject_npf_exit)(struct kvm_vcpu *vcpu,
				       struct x86_exception *fault);

static void kgrr_nested_svm_init_mmu_context(struct kvm_vcpu *vcpu)
{
	WARN_ON(mmu_is_nested(vcpu));
	(*kgre_kvm_init_shadow_mmu)(vcpu);
	vcpu->arch.mmu.set_cr3           = (*kgre_nested_svm_set_tdp_cr3);
	vcpu->arch.mmu.get_cr3           = (*kgre_nested_svm_get_tdp_cr3);
	vcpu->arch.mmu.get_pdptr         = (*kgre_nested_svm_get_tdp_pdptr);
	vcpu->arch.mmu.inject_page_fault = (*kgre_nested_svm_inject_npf_exit);
	vcpu->arch.mmu.shadow_root_level = get_npt_level();
	(*kgre_reset_shadow_zero_bits_mask)(vcpu, &vcpu->arch.mmu);
	vcpu->arch.walk_mmu              = &vcpu->arch.nested_mmu;
}

static int (*kgre_nested_svm_check_permissions)(struct vcpu_svm *svm);

static void *(*kgre_nested_svm_map)(struct vcpu_svm *svm, u64 gpa, struct page **_page);

static void kgrr_nested_svm_unmap(struct page *page)
{
	kunmap(page);
	(*kgre_kvm_release_page_dirty)(page);
}

static inline void copy_vmcb_control_area(struct vmcb *dst_vmcb, struct vmcb *from_vmcb)
{
	struct vmcb_control_area *dst  = &dst_vmcb->control;
	struct vmcb_control_area *from = &from_vmcb->control;

	dst->intercept_cr         = from->intercept_cr;
	dst->intercept_dr         = from->intercept_dr;
	dst->intercept_exceptions = from->intercept_exceptions;
	dst->intercept            = from->intercept;
	dst->iopm_base_pa         = from->iopm_base_pa;
	dst->msrpm_base_pa        = from->msrpm_base_pa;
	dst->tsc_offset           = from->tsc_offset;
	dst->asid                 = from->asid;
	dst->tlb_ctl              = from->tlb_ctl;
	dst->int_ctl              = from->int_ctl;
	dst->int_vector           = from->int_vector;
	dst->int_state            = from->int_state;
	dst->exit_code            = from->exit_code;
	dst->exit_code_hi         = from->exit_code_hi;
	dst->exit_info_1          = from->exit_info_1;
	dst->exit_info_2          = from->exit_info_2;
	dst->exit_int_info        = from->exit_int_info;
	dst->exit_int_info_err    = from->exit_int_info_err;
	dst->nested_ctl           = from->nested_ctl;
	dst->event_inj            = from->event_inj;
	dst->event_inj_err        = from->event_inj_err;
	dst->nested_cr3           = from->nested_cr3;
	dst->lbr_ctl              = from->lbr_ctl;
}

static int (*kgre_nested_svm_vmexit)(struct vcpu_svm *svm);

static bool kgrr_nested_svm_vmrun_msrpm(struct vcpu_svm *svm)
{
	/*
	 * This function merges the msr permission bitmaps of kvm and the
	 * nested vmcb. It is optimized in that it only merges the parts where
	 * the kvm msr permission bitmap may contain zero bits
	 */
	int i;

	if (!(svm->nested.intercept & (1ULL << INTERCEPT_MSR_PROT)))
		return true;

	for (i = 0; i < MSRPM_OFFSETS; i++) {
		u32 value, p;
		u64 offset;

		if ((*kgre_msrpm_offsets)[i] == 0xffffffff)
			break;

		p      = (*kgre_msrpm_offsets)[i];
		offset = svm->nested.vmcb_msrpm + (p * 4);

		if ((*kgre_kvm_vcpu_read_guest)(&svm->vcpu, offset, &value, 4))
			return false;

		svm->nested.msrpm[p] = svm->msrpm[p] | value;
	}

	svm->vmcb->control.msrpm_base_pa = __pa(svm->nested.msrpm);

	return true;
}

static bool kgrr_nested_vmcb_checks(struct vmcb *vmcb)
{
	if ((vmcb->control.intercept & (1ULL << INTERCEPT_VMRUN)) == 0)
		return false;

	if (vmcb->control.asid == 0)
		return false;

	if (vmcb->control.nested_ctl && !(*kgre_npt_enabled))
		return false;

	return true;
}

/*
 * Fix CVE-2021-3653
 *  +1 line
 */
#define V_IRQ_INJECTION_BITS_MASK (V_IRQ_MASK | V_INTR_PRIO_MASK | V_IGN_TPR_MASK)

static bool kgrp_nested_svm_vmrun(struct vcpu_svm *svm)
{
	struct vmcb *nested_vmcb;
	struct vmcb *hsave = svm->nested.hsave;
	struct vmcb *vmcb = svm->vmcb;
	struct page *page;
	u64 vmcb_gpa;

	vmcb_gpa = svm->vmcb->save.rax;

	nested_vmcb = (*kgre_nested_svm_map)(svm, svm->vmcb->save.rax, &page);
	if (!nested_vmcb)
		return false;

	if (!kgrr_nested_vmcb_checks(nested_vmcb)) {
		nested_vmcb->control.exit_code    = SVM_EXIT_ERR;
		nested_vmcb->control.exit_code_hi = 0;
		nested_vmcb->control.exit_info_1  = 0;
		nested_vmcb->control.exit_info_2  = 0;

		kgrr_nested_svm_unmap(page);

		return false;
	}

	kgrr_trace_kvm_nested_vmrun(svm->vmcb->save.rip, vmcb_gpa,
			       nested_vmcb->save.rip,
			       nested_vmcb->control.int_ctl,
			       nested_vmcb->control.event_inj,
			       nested_vmcb->control.nested_ctl);

	kgrr_trace_kvm_nested_intercepts(nested_vmcb->control.intercept_cr & 0xffff,
				    nested_vmcb->control.intercept_cr >> 16,
				    nested_vmcb->control.intercept_exceptions,
				    nested_vmcb->control.intercept);

	/* Clear internal status */
	kvm_clear_exception_queue(&svm->vcpu);
	kvm_clear_interrupt_queue(&svm->vcpu);

	/*
	 * Save the old vmcb, so we don't need to pick what we save, but can
	 * restore everything when a VMEXIT occurs
	 */
	hsave->save.es     = vmcb->save.es;
	hsave->save.cs     = vmcb->save.cs;
	hsave->save.ss     = vmcb->save.ss;
	hsave->save.ds     = vmcb->save.ds;
	hsave->save.gdtr   = vmcb->save.gdtr;
	hsave->save.idtr   = vmcb->save.idtr;
	hsave->save.efer   = svm->vcpu.arch.efer;
	hsave->save.cr0    = kgrr_kvm_read_cr0(&svm->vcpu);
	hsave->save.cr4    = svm->vcpu.arch.cr4;
	hsave->save.rflags = (*kgre_kvm_get_rflags)(&svm->vcpu);
	hsave->save.rip    = kgrr_kvm_rip_read(&svm->vcpu);
	hsave->save.rsp    = vmcb->save.rsp;
	hsave->save.rax    = vmcb->save.rax;
	if ((*kgre_npt_enabled))
		hsave->save.cr3    = vmcb->save.cr3;
	else
		hsave->save.cr3    = kgrr_kvm_read_cr3(&svm->vcpu);

	copy_vmcb_control_area(hsave, vmcb);

	if ((*kgre_kvm_get_rflags)(&svm->vcpu) & X86_EFLAGS_IF)
		svm->vcpu.arch.hflags |= HF_HIF_MASK;
	else
		svm->vcpu.arch.hflags &= ~HF_HIF_MASK;

	if (nested_vmcb->control.nested_ctl) {
		(*kgre_kvm_mmu_unload)(&svm->vcpu);
		svm->nested.nested_cr3 = nested_vmcb->control.nested_cr3;
		kgrr_nested_svm_init_mmu_context(&svm->vcpu);
	}

	/* Load the nested guest state */
	svm->vmcb->save.es = nested_vmcb->save.es;
	svm->vmcb->save.cs = nested_vmcb->save.cs;
	svm->vmcb->save.ss = nested_vmcb->save.ss;
	svm->vmcb->save.ds = nested_vmcb->save.ds;
	svm->vmcb->save.gdtr = nested_vmcb->save.gdtr;
	svm->vmcb->save.idtr = nested_vmcb->save.idtr;
	(*kgre_kvm_set_rflags)(&svm->vcpu, nested_vmcb->save.rflags);
	(*kgre_svm_set_efer)(&svm->vcpu, nested_vmcb->save.efer);
	(*kgre_svm_set_cr0)(&svm->vcpu, nested_vmcb->save.cr0);
	(*kgre_svm_set_cr4)(&svm->vcpu, nested_vmcb->save.cr4);
	if ((*kgre_npt_enabled)) {
		svm->vmcb->save.cr3 = nested_vmcb->save.cr3;
		svm->vcpu.arch.cr3 = nested_vmcb->save.cr3;
	} else
		(void)(*kgre_kvm_set_cr3)(&svm->vcpu, nested_vmcb->save.cr3);

	/* Guest paging mode is active - reset mmu */
	(*kgre_kvm_mmu_reset_context)(&svm->vcpu);

	svm->vmcb->save.cr2 = svm->vcpu.arch.cr2 = nested_vmcb->save.cr2;
	kvm_register_write(&svm->vcpu, VCPU_REGS_RAX, nested_vmcb->save.rax);
	kvm_register_write(&svm->vcpu, VCPU_REGS_RSP, nested_vmcb->save.rsp);
	kvm_register_write(&svm->vcpu, VCPU_REGS_RIP, nested_vmcb->save.rip);

	/* In case we don't even reach vcpu_run, the fields are not updated */
	svm->vmcb->save.rax = nested_vmcb->save.rax;
	svm->vmcb->save.rsp = nested_vmcb->save.rsp;
	svm->vmcb->save.rip = nested_vmcb->save.rip;
	svm->vmcb->save.dr7 = nested_vmcb->save.dr7;
	svm->vmcb->save.dr6 = nested_vmcb->save.dr6;
	svm->vmcb->save.cpl = nested_vmcb->save.cpl;

	svm->nested.vmcb_msrpm = nested_vmcb->control.msrpm_base_pa & ~0x0fffULL;
	svm->nested.vmcb_iopm  = nested_vmcb->control.iopm_base_pa  & ~0x0fffULL;

	/* cache intercepts */
	svm->nested.intercept_cr         = nested_vmcb->control.intercept_cr;
	svm->nested.intercept_dr         = nested_vmcb->control.intercept_dr;
	svm->nested.intercept_exceptions = nested_vmcb->control.intercept_exceptions;
	svm->nested.intercept            = nested_vmcb->control.intercept;

	(*kgre_svm_flush_tlb)(&svm->vcpu);
	/*
	 * Fix CVE-2021-3653
	 *  -1 line, +5 lines
	 */
	svm->vmcb->control.int_ctl = nested_vmcb->control.int_ctl &
				(V_TPR_MASK | V_IRQ_INJECTION_BITS_MASK);

	svm->vmcb->control.int_ctl |= V_INTR_MASKING_MASK;

	if (nested_vmcb->control.int_ctl & V_INTR_MASKING_MASK)
		svm->vcpu.arch.hflags |= HF_VINTR_MASK;
	else
		svm->vcpu.arch.hflags &= ~HF_VINTR_MASK;

	if (svm->vcpu.arch.hflags & HF_VINTR_MASK) {
		/* We only want the cr8 intercept bits of the guest */
		kgrr_clr_cr_intercept(svm, INTERCEPT_CR8_READ);
		kgrr_clr_cr_intercept(svm, INTERCEPT_CR8_WRITE);
	}

	/* We don't want to see VMMCALLs from a nested guest */
	kgrr_clr_intercept(svm, INTERCEPT_VMMCALL);

	svm->vmcb->control.lbr_ctl = nested_vmcb->control.lbr_ctl;
	svm->vmcb->control.int_vector = nested_vmcb->control.int_vector;
	svm->vmcb->control.int_state = nested_vmcb->control.int_state;
	svm->vmcb->control.tsc_offset += nested_vmcb->control.tsc_offset;
	svm->vmcb->control.event_inj = nested_vmcb->control.event_inj;
	svm->vmcb->control.event_inj_err = nested_vmcb->control.event_inj_err;

	kgrr_nested_svm_unmap(page);

	/* Enter Guest-Mode */
	enter_guest_mode(&svm->vcpu);

	/*
	 * Merge guest and host intercepts - must be called  with vcpu in
	 * guest-mode to take affect here
	 */
	(*kgre_recalc_intercepts)(svm);

	svm->nested.vmcb = vmcb_gpa;

	enable_gif(svm);

	mark_all_dirty(svm->vmcb);

	return true;
}

int kgrp_vmrun_interception(struct vcpu_svm *svm)
{
	if ((*kgre_nested_svm_check_permissions)(svm))
		return 1;

	/* Save rip after vmrun instruction */
	kvm_rip_write(&svm->vcpu, kgrr_kvm_rip_read(&svm->vcpu) + 3);

	if (!kgrp_nested_svm_vmrun(svm))
		return 1;

	if (!kgrr_nested_svm_vmrun_msrpm(svm))
		goto failed;

	return 1;

failed:

	svm->vmcb->control.exit_code    = SVM_EXIT_ERR;
	svm->vmcb->control.exit_code_hi = 0;
	svm->vmcb->control.exit_info_1  = 0;
	svm->vmcb->control.exit_info_2  = 0;

	(*kgre_nested_svm_vmexit)(svm);

	return 1;
}



#include <linux/kernel.h>
#include <linux/module.h>
#include "kgr_patch_bsc1189420.h"
#include "../kallsyms_relocs.h"

#define KGR_PATCHED_MODULE "kvm_amd"

static struct kgr_kallsyms_reloc kgr_funcs[] = {
	{ "npt_enabled", (void *)&kgre_npt_enabled, "kvm_amd" },
	{ "__tracepoint_kvm_nested_intercepts",
	  (void *)&kgre___tracepoint_kvm_nested_intercepts, "kvm" },
	{ "__tracepoint_kvm_nested_vmrun",
	  (void *)&kgre___tracepoint_kvm_nested_vmrun, "kvm" },
	{ "msrpm_offsets", (void *)&kgre_msrpm_offsets, "kvm_amd" },
	{ "kvm_x86_ops", (void *)&kgre_kvm_x86_ops, "kvm" },
	{ "nested_svm_vmexit", (void *)&kgre_nested_svm_vmexit, "kvm_amd" },
	{ "svm_set_cr0", (void *)&kgre_svm_set_cr0, "kvm_amd" },
	{ "svm_flush_tlb", (void *)&kgre_svm_flush_tlb, "kvm_amd" },
	{ "recalc_intercepts", (void *)&kgre_recalc_intercepts, "kvm_amd" },
	{ "svm_set_efer", (void *)&kgre_svm_set_efer, "kvm_amd" },
	{ "svm_set_cr4", (void *)&kgre_svm_set_cr4, "kvm_amd" },
	{ "nested_svm_get_tdp_cr3", (void *)&kgre_nested_svm_get_tdp_cr3,
	  "kvm_amd" },
	{ "nested_svm_get_tdp_pdptr", (void *)&kgre_nested_svm_get_tdp_pdptr,
	  "kvm_amd" },
	{ "nested_svm_set_tdp_cr3", (void *)&kgre_nested_svm_set_tdp_cr3,
	  "kvm_amd" },
	{ "nested_svm_inject_npf_exit",
	  (void *)&kgre_nested_svm_inject_npf_exit, "kvm_amd" },
	{ "nested_svm_check_permissions",
	  (void *)&kgre_nested_svm_check_permissions, "kvm_amd" },
	{ "nested_svm_map", (void *)&kgre_nested_svm_map, "kvm_amd" },
	{ "kvm_mmu_reset_context", (void *)&kgre_kvm_mmu_reset_context, "kvm" },
	{ "kvm_set_cr3", (void *)&kgre_kvm_set_cr3, "kvm" },
	{ "kvm_get_rflags", (void *)&kgre_kvm_get_rflags, "kvm" },
	{ "kvm_set_rflags", (void *)&kgre_kvm_set_rflags, "kvm" },
	{ "kvm_mmu_unload", (void *)&kgre_kvm_mmu_unload, "kvm" },
	{ "kvm_release_page_dirty", (void *)&kgre_kvm_release_page_dirty,
	  "kvm" },
	{ "kvm_vcpu_read_guest", (void *)&kgre_kvm_vcpu_read_guest, "kvm" },
	{ "reset_shadow_zero_bits_mask",
	  (void *)&kgre_reset_shadow_zero_bits_mask, "kvm" },
	{ "kvm_init_shadow_mmu", (void *)&kgre_kvm_init_shadow_mmu, "kvm" },
};

static int kgr_patch_bsc1189420_module_notify(struct notifier_block *nb,
					      unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = __kgr_resolve_kallsyms_relocs(kgr_funcs, ARRAY_SIZE(kgr_funcs));
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc1189420_module_nb = {
	.notifier_call = kgr_patch_bsc1189420_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc1189420_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = __kgr_resolve_kallsyms_relocs(kgr_funcs,
						    ARRAY_SIZE(kgr_funcs));
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc1189420_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc1189420_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc1189420_module_nb);
}

#endif /* IS_ENABLED(CONFIG_KVM_AMD) */
